!> triangular mesh 
module o_mesh

  implicit none


  !> boundary edge, from input and the curved representation, if any
  type, public :: mesh
     integer :: nelem    ! number of elements
     integer :: npoin    ! number of nodes
     integer :: ipoin    ! number of interior nodes
     integer :: nbelm    ! number of boundary elements
     integer :: nbc      ! number of boundary components

     real, allocatable, dimension(:,:) :: xp   ! coordinates of nodes
     integer, allocatable, dimension(:,:) :: nodes ! indexes of vertices of elements
     integer, allocatable, dimension(:,:) :: lbn   ! indexes of boundary edges
     integer, allocatable, dimension(:,:) :: ip ! interior or boundary nodes?
     real, allocatable, dimension(:) :: area  ! areas of elements
     real, allocatable, dimension(:) :: b_edge  ! only for Neumann nodes
  end type mesh


  public :: Read_mesh


contains

  ! readin of the mesh from the file and setting of arrays
  subroutine Read_mesh(trifile, grid)
    character(len=*), intent( in) :: trifile
    type(mesh), target, intent(inout) :: grid
    real :: r1, r2, r3, r4
    integer, dimension(:,:), pointer :: ip
    integer :: i, i1,i2,i3,i4, itri
    integer :: npoin, nelem, nbelm, nbc, ipoin
    real :: val


    ! reading of mesh
    itri = 10
    open(itri, file=trifile, status='OLD')

    read(itri,*)  npoin, nelem, nbelm, grid%nbc
    grid%npoin = npoin
    grid%nelem = nelem
    grid%nbelm = nbelm

    allocate(grid%xp(1:npoin, 1:2) )
    allocate(grid%nodes(1:nelem, 1:3) )
    allocate(grid%lbn(1:nbelm, 1:3) )
    allocate(grid%ip(1:npoin, 1:2) )

    allocate(grid%area(1:nelem), source = 0.0 ) ! computed in subroutine STIFF_MATRIX


    ! skipping of data
    read(itri, *) r1, r2, i1, i2, r3, r4, i3, i4

    ! coordinates of vertices
    do i=1,npoin
       read(itri, *) grid%xp(i, 1:2)
    enddo

    ! indexes of vertices of elements 
    do i=1,nelem
       read(itri, *) grid%nodes(i, 1:3)
       !print*,'nodes:',i, nodes(i, 1:3)
    enddo

    ! indexes of vertices of boundary edges + index of component
    do i=1,nbelm
       read(itri, *) grid%lbn(i, 1:3)
    enddo


    ip => grid%ip(1:npoin, 1:2)

    ! setting of array ip: ip(i) == 1 the -th node is interior
    ! ip(i, 1) == 1  => interior node
    ! ip(i, 1) == 0  => node on the Neumann part of boundary 
    ! ip(i, 1) == -1  => node on the Dirichlet part of boundary

    ip(:,1) = 1
    ip(:,2) = 0
    ! first all boundary nodes
    do i=1,nbelm
       ! i1 and i2 ... verices of an boundary edge
       i1 = grid%lbn(i,1)
       i2 = grid%lbn(i,2)

       ip(i1, 1) = 0  ! Boundary node
       ip(i2, 1) = 0
    enddo

    ! only Dirichlet nodes
    do i=1,nbelm
       ! i1 and i2 ... verices of an boundary edge
       i1 = grid%lbn(i,1)
       i2 = grid%lbn(i,2)
       i3 = grid%lbn(i,3)  ! number of boundary component

       ! if i3 == 2 then Neumann BC, otherwise Dirichlet
       !if( i3 /= 2) then
       if( i3 /= -2) then   ! NO NEUMANN BC
          ip(i1, 1) =  - 1  ! Boundary node
          ip(i2, 1) = - 1
       endif
    enddo

    ! setting the number of interior and Neumann nodes = number of unknowns
    ipoin = 0
    do i=1,npoin

       if(ip(i, 1) >= 0) then   ! Neumann or interior
          ipoin = ipoin + 1
          ip(i, 2) = ipoin
       endif

    enddo

    grid%ipoin = ipoin
    print*,"Number of interior or Neumann nodes = ", ipoin

    close(itri)

    ! only for Neumann nodes mean value of lengths of adjacent edges
    allocate(grid%b_edge(1:npoin), source = 0.0 )
    do i=1,nbelm
       ! i1 and i2 ... verices of an boundary edge
       i1 = grid%lbn(i,1)
       i2 = grid%lbn(i,2)

       ! 1/2 of the length of the edge
       val = norm2( grid%xp(i1, 1:2)  - grid%xp(i2, 1:2) ) / 2

       if(ip(i1, 1) == 0)  grid%b_edge(i1) = grid%b_edge(i1) + val
       if(ip(i2, 1) == 0)  grid%b_edge(i2) = grid%b_edge(i2) + val
    enddo


  end subroutine Read_mesh



end module o_mesh



