!> module for the operation with FE solution 
module o_sol
  use o_mesh

  
  implicit none

contains

  ! plot of the P1 FE solution to a file suitbale for gnuplot
  subroutine Plot_solution(grid,  solfile, u )
    type(mesh), target, intent(inout) :: grid
   character(len=*), intent( in) :: solfile
   real, dimension(1:grid%npoin), intent(inout) :: u  ! solution in nodes
    real, dimension(:,:), pointer :: xp  ! nodes coordinates
    integer, dimension(:,:), pointer :: nodes  ! index vertex
    integer :: i,j,k, l, npoin, nelem

    npoin = grid%npoin
    nelem = grid%nelem

    nodes => grid%nodes(1:nelem,1:3)

    xp => grid%xp(1:npoin,1:2)

    open(90, file=solfile, status='UNKNOWN')

  ! drawing of matrix structure
    do i=1,grid%nelem

       write(90, *) xp(nodes(i, 1), 1:2), u(nodes(i,1))
       write(90, *) xp(nodes(i, 2), 1:2), u(nodes(i,2))
       write(90, *) xp(nodes(i, 3), 1:2), u(nodes(i,3))
       write(90, *) xp(nodes(i, 1), 1:2), u(nodes(i,1))
       write(90, '(x)' )
       write(90, '(x)' )
       write(90, '(x)' )
    enddo

    close(90)
    
  end subroutine Plot_solution

  ! setting of the RHS for the linear algebraci problem
  ! source term + terms from non-homogeous Dirichlet BC
  subroutine InitSol_RHS(grid, mS, u,  v,f)
    type(mesh), target, intent(inout) :: grid
    real, dimension(1:grid%npoin, 1:grid%npoin), intent(inout) :: mS  ! stiff matrix
    real, dimension(1:grid%npoin), intent(inout) :: u  ! solution in nodes
    real, dimension(1:grid%ipoin), intent(inout) :: v  ! init solution in nodes
    real, dimension(1:grid%ipoin), intent(inout) :: f  ! RHS in interior nodes
    real, dimension(:,:), pointer :: xp  ! nodes coordinates
    real, dimension(:), pointer :: area  ! area of elements
    integer, dimension(:,:), pointer :: nodes  ! vertex indexes
    integer, dimension(:,:), pointer :: ip  ! interior / boundary
    real, dimension(:), allocatable :: xi  ! coordinates in middle of edges
    integer :: i,j, j1, k, l, npoin, ipoin, nelem, k1, k2
    real :: val

    allocate(xi(1:2) )
    
    npoin = grid%npoin
    ipoin = grid%ipoin
    nelem = grid%nelem

    ip => grid%ip(1:npoin,1:2)

    xp => grid%xp(1:npoin,1:2)

    nodes => grid%nodes(1:nelem,1:3)

    area => grid%area(1:nelem)

    
    u = 0.

    ! setting of the Dicihlet BC
    do i=1, npoin
       if(ip(i, 1) < 0) then
          u(i) = set_DBC(xp(i, 1), xp(i,2) )
       else
          ! initial setting of the solution in inner and Neumann nodes
          v(ip(i,2) ) = u(i) ! usually =0
       endif
    enddo

    ! setting of source term
    f(:) = 0
    
    do i=1, nelem
       do j=1, 3
          j1 = mod(j, 3) + 1

          k1 = nodes(i, j)
          k2 = nodes(i, j1)
          xi(1:2) = ( xp( k1, 1:2) + xp(k2, 1:2) ) /2

          val = set_RHS( xi(1), xi(2) )
          
          if(ip(k1, 1) == 1) f(ip(k1,2)) = f(ip(k1,2)) + val * area(i) / 6
          if(ip(k2, 1) == 1) f(ip(k2,2)) = f(ip(k2,2)) + val * area(i) / 6
       enddo
       
    enddo


    ! modification  of the right-hand side by the Dirichlet BC
    do i=1, npoin
       if(ip(i, 1) >= 0) then ! interior node, the equation
          k = ip(i,2)

          do j=1, npoin
             if(ip(j, 1) < 0) then  ! Dirichlet  node (the data)
                
                f(k) = f(k) - mS(i, j) * u(j )   ! u already contains Dirichlet data
             endif
          enddo
       endif
    end do


    ! modification  of the right-hand side by the Neumann BC
    do i=1, npoin
       if(ip(i, 1) == 0) then ! Neumann  node, the equation
          k = ip(i,2)

          val = set_NBC(xp(k, 1), xp(k,2) )
          
          f(k) = f(k) + val * grid%b_edge(i) 
       endif
    end do

    
    deallocate(xi )
    
  end subroutine InitSol_RHS


  ! setting of the Dirichlet boundary condition at x,y
  function set_DBC(x, y) result( fun )
    real, intent(in) :: x, y  ! point coordinates
    real :: fun

    fun = 0.
    !fun = x**2 + y**2
  end function set_DBC


  ! setting of the Neumann boundary condition at x,y
  function set_NBC(x, y) result( fun )
    real, intent(in) :: x, y  ! point coordinates
    real :: fun

    fun = 0.
    fun = 1.

  end function set_NBC


  ! evaluation of the RHS of the equation at x,y
  function set_RHS(x, y) result( fun )
    real, intent(in) :: x, y  ! point coordinates
    real :: fun

    fun = 2.
    !fun = 2*y*(1-y) + 2*x*(1-x)
    !fun = -4
    
  end function set_RHS

end module o_sol


